from __future__ import (absolute_import, division, unicode_literals)
from math import log10, floor


def format_numbers(*args):
    scales = tuple(int(round(log10(abs(x)))) if x != 0 else 1 for x in args)
    nmax = max(scales)
    nsig = min(x for x in scales if nmax - x < 9) - 1
    all_pos = True
    for arg in args:
        if arg < 0:
            all_pos = False
            break
    if nmax <= 3 and nsig >= -3:
        if nsig >= 0:
            return tuple(("%i" if x < 0 or all_pos else " %i") % round(x, -nsig) for x in args)
        else:
            return tuple((("%%.%if" if x < 0 or all_pos else " %%.%if") % min(-nsig, 3)) % x for x in args)
    result = [("%%.%if" if x < 0 or all_pos else " %%.%if") % min(nmax-nsig, 3) % (x / 10 ** nmax) for x in args]
    if nmax != 0:
        return tuple(x + 'E%i' % nmax for x in result)
    return tuple(result)


def format_row(widths, *args):
    return "".join(('|{0:^%i}' % (w-1) if w > 0 else '| {0:%i}' % (-w-2)).format(a) for (w, a) in zip(widths, args)) + "|"


def fmin(fmin):
    goaledm = 1e-4 * fmin.tolerance * fmin.up
    # despite what the doc said the code is actually 1e-4
    # http://wwwasdoc.web.cern.ch/wwwasdoc/hbook_html3/node125.html

    ws = (-32, 33)
    i1 = format_row(ws, 'FCN = %.4G' % fmin.fval,
        'Ncalls=%i (%i total)' % (fmin.nfcn, fmin.ncalls))
    i2 = format_row(ws, 'EDM = %.3G (Goal: %G)' % (fmin.edm, goaledm),
        'up = %.1f' % fmin.up)
    ws = (16, 16, 12, 21)
    h1 = format_row(ws, "Valid Min.", "Valid Param.", "Above EDM", "Reached call limit")
    v1 = format_row(ws, repr(fmin.is_valid), repr(fmin.has_valid_parameters),
                         repr(fmin.is_above_max_edm), repr(fmin.has_reached_call_limit))
    ws = (16, 16, 12, 12, 9)
    h2 = format_row(ws, "Hesse failed", "Has cov.", "Accurate", "Pos. def.", "Forced")
    v2 = format_row(ws, repr(fmin.hesse_failed), repr(fmin.has_covariance),
        repr(fmin.has_accurate_covar), repr(fmin.has_posdef_covar),
        repr(fmin.has_made_posdef_covar))
    
    l = len(h1) * "-"
    return "\n".join((l, i1, i2, l, h1, l, v1, l, h2, l, v2, l))


def params(mps):
    vnames = [mp.name for mp in mps]
    name_width = max([4] + [len(x) for x in vnames])
    num_width = max(2, len("%i" % (len(vnames) - 1)))

    ws = (-num_width - 2, -name_width - 3, 12, 12, 13, 13, 10, 10, 8)
    h = format_row(ws, '', 'Name', 'Value', 'Hesse Err',
        "Minos Err-", "Minos Err+", "Limit-", "Limit+", "Fixed")
    l = '-' * len(h)
    lines = [l, h, l]
    mes = mps.merrors
    for i, mp in enumerate(mps):
        if mes and mp.name in mes:
            me = mes[mp.name]
            val, err, mel, meu = format_numbers(mp.value, mp.error, me.lower, me.upper)
        else:
            val, err = format_numbers(mp.value, mp.error)
            mel = ''
            meu = ''
        lines.append(format_row(ws,
            str(i), mp.name,
            val,
            err,
            mel,
            meu,
            '%g' % mp.lower_limit if mp.lower_limit is not None else '',
            '%g' % mp.upper_limit if mp.upper_limit is not None else '',
            'yes' if mp.is_fixed else 'CONST' if mp.is_const else ''
        ))
    lines.append(l)
    return "\n".join(lines)


def merror(me):
    mel, meu = format_numbers(me.lower, me.upper)
    stat = 'Valid' if me.is_valid else 'Invalid'
    summary = '| {0:^15s} | {1:^27s} |'.format(me.name, stat)
    error = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
        'Error', mel, meu)
    valid = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
        'Valid', str(me.lower_valid), str(me.upper_valid))
    at_limit = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
        'At Limit', str(me.at_lower_limit), str(me.at_upper_limit))
    max_fcn = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
        'Max FCN', str(me.at_lower_max_fcn), str(me.at_upper_max_fcn))
    new_min = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
        'New Min', str(me.lower_new_min), str(me.upper_new_min))
    hline = '-' * len(error)
    return "\n".join((hline, summary, hline, error, valid,
                      at_limit, max_fcn, new_min, hline))


def matrix(m):
    n = len(m)

    is_correlation = True
    for i in range(n):
        if m[i][i] != 1.0:
            is_correlation = False
            break

    args = []
    for mi in m:
        for mj in mi:
            args.append(mj)
    nums = format_numbers(*args)        

    def row_fmt(args):
        s = '| ' + args[0] + ' |'
        for x in args[1:]:
            s += ' ' + x
        s += ' |'
        return s

    first_row_width = max(len(v) for v in m.names)
    row_width = max(first_row_width, max(len(v) for v in nums))
    v_names = [('{:>%is}' % first_row_width).format(x) for x in m.names]
    h_names = [('{:>%is}' % row_width).format(x) for x in m.names]
    val_fmt = ('{:>%is}' % row_width).format

    header = row_fmt([' ' * first_row_width] + h_names)
    hline = '-' * len(header)
    lines = [hline, header, hline]
    
    for i, vn in enumerate(v_names):
        lines.append(row_fmt([vn] + [val_fmt(nums[n*i + j]) for j in range(n)]))
    lines.append(hline)
    return "\n".join(lines)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from numpy.testing import assert_allclose
import os

__all__ = [
    'requires_dependency',
    'assert_allclose',
]


def requires_dependency(*names):
    """Decorator to declare required dependencies for tests.

    Parameters
    ----------
    name : str
        Package name, e.g. 'numpy' or 'ipython'

    Examples
    --------

    ::

        from iminuit.tests.utils import requires_dependency

        @requires_dependency('numpy')
        def test_using_numpy():
            import numpy
            ...
    """
    skip_it = False
    for name in names:
        path = os.path.dirname(__file__)
        p = os.path.join(path, name + ".pyx")
        if os.path.exists(p):
            continue
        try:
            __import__(name)
        except ImportError:
            skip_it = True

    reason = 'Missing dependency: {}'.format(name)
    import pytest
    return pytest.mark.skipif(skip_it, reason=reason)



from __future__ import (absolute_import, division, print_function)
# Here we don't import unicode_literals, because otherwise we cannot write
# in "\usepackage" and similar in the string literals below.
# The problem is an inconsistency regarding the effect of
# from __future__ import unicode_literals in Python 2 and 3, which is
# explained here:
# https://stackoverflow.com/questions/7602171/unicode-error-unicodeescape-codec-cant-decode-bytes-string-with-u
# We want the same code to work in Python 2 and 3 and chose this solution.
from iminuit import Minuit
from iminuit.util import Params, Param, Matrix
from iminuit import repr_html, repr_text
import pytest


def f1(x, y):
    return (x - 2) ** 2 + (y - 1) ** 2 / 0.25 + 1


def test_html_tag():
    s = repr_html.Html()
    with repr_html.Tag("foo", s):
        s += "bar"
    assert str(s) == "<foo>\nbar\n</foo>\n"


def test_format_numbers():
    assert repr_text.format_numbers(1.2567, 0.1234) == ("1.26", "0.12")
    assert repr_text.format_numbers(1.2567e3, 0.1234e3) == ("1260", "120")
    assert repr_text.format_numbers(1.2567e4, 0.1234e4) == ("1.26E4", "0.12E4")
    assert repr_text.format_numbers(1.2567e-1, 0.1234e-1) == ("0.126", "0.012")
    assert repr_text.format_numbers(1.2567e-2, 0.1234e-2) == ("1.26E-2", "0.12E-2")
    assert repr_text.format_numbers(1.0, 0.0, 0.25) == ("1.00", "0.00", "0.25")
    assert repr_text.format_numbers(0, 1, -1) == (" 0.0", " 1.0", "-1.0")
    assert repr_text.format_numbers(2, -1, 1) == (" 2.0", "-1.0", " 1.0")
    assert repr_text.format_numbers(2.01, -1.01, 1.01) == (" 2.0", "-1.0", " 1.0")
    assert repr_text.format_numbers(1.999, -0.999, 0.999) == (" 2.0", "-1.0", " 1.0")
    assert repr_text.format_numbers(1, 0.5, -0.5) == (" 1.0", " 0.5", "-0.5")
    assert repr_text.format_numbers(1.0, 1e-10) == ("1.0", "0.0")
    assert repr_text.format_numbers(1.234567e11, -1.234567e-11) == (" 1.2E11", "-0.0E11")


@pytest.fixture
def minuit():
    m = Minuit(f1, x=0, y=0, pedantic=False, print_level=0)
    m.tol = 1e-4
    m.migrad()
    m.hesse()
    m.minos()
    return m


def test_html_fmin(minuit):
    fmin = minuit.get_fmin()
    assert r"""<table>
<tr>
<td colspan="2" title="Minimum value of function">
FCN = 1
</td>
<td align="center" colspan="3" title="No. of calls in last algorithm and total number of calls">
Ncalls = 24 (67 total)
</td>
</tr>
<tr>
<td colspan="2" title="Estimated distance to minimum and target threshold">
EDM = %.3G (Goal: 1E-08)
</td>
<td align="center" colspan="3" title="Increase in FCN which corresponds to 1 standard deviation">
up = 1.0
</td>
</tr>
<tr>
<td align="center" title="Validity of the migrad call">
Valid Min.
</td>
<td align="center" title="Validity of parameters">
Valid Param.
</td>
<td align="center" title="Is EDM above goal EDM?">
Above EDM
</td>
<td align="center" colspan="2" title="Did last migrad call reach max call limit?">
Reached call limit
</td>
</tr>
<tr>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
False
</td>
<td align="center" colspan="2" style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td align="center" title="Did Hesse fail?">
Hesse failed
</td>
<td align="center" title="Has covariance matrix">
Has cov.
</td>
<td align="center" title="Is covariance matrix accurate?">
Accurate
</td>
<td align="center" title="Is covariance matrix positive definite?">
Pos. def.
</td>
<td align="center" title="Was positive definiteness enforced by Minuit?">
Forced
</td>
</tr>
<tr>
<td align="center" style="background-color:#92CCA6;">
False
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
True
</td>
<td align="center" style="background-color:#92CCA6;">
False
</td>
</tr>
</table>
""" % fmin.edm == fmin._repr_html_()


def test_html_params(minuit):
    assert r"""<table>
<tr style="background-color:#F4F4F4;">
<td/>
<th title="Variable name">
Name
</th>
<th title="Value of parameter">
Value
</th>
<th title="Hesse error">
Hesse Error
</th>
<th title="Minos lower error">
Minos Error-
</th>
<th title="Minos upper error">
Minos Error+
</th>
<th title="Lower limit of the parameter">
Limit-
</th>
<th title="Upper limit of the parameter">
Limit+
</th>
<th title="Is the parameter fixed in the fit">
Fixed
</th>
</tr>
<tr style="background-color:#FFFFFF;">
<td>
0
</td>
<td>
x
</td>
<td>
0.0
</td>
<td>
1.0
</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
<tr style="background-color:#F4F4F4;">
<td>
1
</td>
<td>
y
</td>
<td>
0.0
</td>
<td>
1.0
</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
</table>
""" == minuit.get_initial_param_states()._repr_html_()

    assert """<table>
<tr style="background-color:#F4F4F4;">
<td/>
<th title="Variable name">
Name
</th>
<th title="Value of parameter">
Value
</th>
<th title="Hesse error">
Hesse Error
</th>
<th title="Minos lower error">
Minos Error-
</th>
<th title="Minos upper error">
Minos Error+
</th>
<th title="Lower limit of the parameter">
Limit-
</th>
<th title="Upper limit of the parameter">
Limit+
</th>
<th title="Is the parameter fixed in the fit">
Fixed
</th>
</tr>
<tr style="background-color:#FFFFFF;">
<td>
0
</td>
<td>
x
</td>
<td>
 2.0
</td>
<td>
 1.0
</td>
<td>
-1.0
</td>
<td>
 1.0
</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
<tr style="background-color:#F4F4F4;">
<td>
1
</td>
<td>
y
</td>
<td>
 1.0
</td>
<td>
 0.5
</td>
<td>
-0.5
</td>
<td>
 0.5
</td>
<td>

</td>
<td>

</td>
<td>

</td>
</tr>
</table>
""" == minuit.get_param_states()._repr_html_()


def test_html_params_with_limits():
    m = Minuit(f1, x=3, y=5, fix_x=True,
               error_x=0.2, error_y=0.1,
               limit_x=(0, None), limit_y=(0, 10),
               errordef=1, print_level=0)
    assert r"""<table>
<tr style="background-color:#F4F4F4;">
<td/>
<th title="Variable name">
Name
</th>
<th title="Value of parameter">
Value
</th>
<th title="Hesse error">
Hesse Error
</th>
<th title="Minos lower error">
Minos Error-
</th>
<th title="Minos upper error">
Minos Error+
</th>
<th title="Lower limit of the parameter">
Limit-
</th>
<th title="Upper limit of the parameter">
Limit+
</th>
<th title="Is the parameter fixed in the fit">
Fixed
</th>
</tr>
<tr style="background-color:#FFFFFF;">
<td>
0
</td>
<td>
x
</td>
<td>
3.00
</td>
<td>
0.20
</td>
<td>

</td>
<td>

</td>
<td>
0
</td>
<td>

</td>
<td>
yes
</td>
</tr>
<tr style="background-color:#F4F4F4;">
<td>
1
</td>
<td>
y
</td>
<td>
5.00
</td>
<td>
0.10
</td>
<td>

</td>
<td>

</td>
<td>
0
</td>
<td>
10
</td>
<td>

</td>
</tr>
</table>
""" == m.get_initial_param_states()._repr_html_()


def test_html_minos(minuit):
    mes = minuit.merrors_struct
    assert r"""<table>
<tr>
<th title="Parameter name">
x
</th>
<td align="center" colspan="2" style="background-color:#92CCA6;">
Valid
</td>
</tr>
<tr>
<td title="Lower and upper minos error of the parameter">
Error
</td>
<td>
-1.0
</td>
<td>
 1.0
</td>
</tr>
<tr>
<td title="Validity of lower/upper minos error">
Valid
</td>
<td style="background-color:#92CCA6;">
True
</td>
<td style="background-color:#92CCA6;">
True
</td>
</tr>
<tr>
<td title="Did scan hit limit of any parameter?">
At Limit
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="Did scan hit function call limit?">
Max FCN
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="New minimum found when doing scan?">
New Min
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
</table>

<table>
<tr>
<th title="Parameter name">
y
</th>
<td align="center" colspan="2" style="background-color:#92CCA6;">
Valid
</td>
</tr>
<tr>
<td title="Lower and upper minos error of the parameter">
Error
</td>
<td>
-0.5
</td>
<td>
 0.5
</td>
</tr>
<tr>
<td title="Validity of lower/upper minos error">
Valid
</td>
<td style="background-color:#92CCA6;">
True
</td>
<td style="background-color:#92CCA6;">
True
</td>
</tr>
<tr>
<td title="Did scan hit limit of any parameter?">
At Limit
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="Did scan hit function call limit?">
Max FCN
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
<tr>
<td title="New minimum found when doing scan?">
New Min
</td>
<td style="background-color:#92CCA6;">
False
</td>
<td style="background-color:#92CCA6;">
False
</td>
</tr>
</table>
""" == mes._repr_html_()


def test_html_matrix(minuit):
    assert r"""<table>
<tr>
<td/>

<th>
x
</th>
<th>
y
</th>
</tr>
<tr>
<th>
x
</th>
<td>
1.00
</td>
<td style="background-color:rgb(163,254,186)">
0.00
</td>
</tr>
<tr>
<th>
y
</th>
<td style="background-color:rgb(163,254,186)">
0.00
</td>
<td>
0.25
</td>
</tr>
</table>
""" == minuit.matrix()._repr_html_()


def test_text_fmin(minuit):
    assert \
r"""------------------------------------------------------------------
| FCN = 1                       |      Ncalls=24 (67 total)      |
| EDM = %.3G (Goal: 1E-08)  |            up = 1.0            |
------------------------------------------------------------------
|  Valid Min.   | Valid Param.  | Above EDM | Reached call limit |
------------------------------------------------------------------
|     True      |     True      |   False   |       False        |
------------------------------------------------------------------
| Hesse failed  |   Has cov.    | Accurate  | Pos. def. | Forced |
------------------------------------------------------------------
|     False     |     True      |   True    |   True    | False  |
------------------------------------------------------------------""" \
    % minuit.get_fmin().edm == str(minuit.get_fmin())


def test_text_params(minuit):
    assert \
r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |    0.0    |    1.0    |            |            |         |         |       |
| 1 | y    |    0.0    |    1.0    |            |            |         |         |       |
------------------------------------------------------------------------------------------""" \
    == str(minuit.get_initial_param_states())

    assert \
r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |    2.0    |    1.0    |    -1.0    |     1.0    |         |         |       |
| 1 | y    |    1.0    |    0.5    |    -0.5    |     0.5    |         |         |       |
------------------------------------------------------------------------------------------""" \
    == str(minuit.get_param_states())


def test_text_params_with_limits():
    m = Minuit(f1, x=3, y=5, fix_x=True,
               error_x=0.2, error_y=0.1,
               limit_x=(0, None), limit_y=(0, 10),
               errordef=1)
    assert \
r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    |   3.00    |   0.20    |            |            |    0    |         |  yes  |
| 1 | y    |   5.00    |   0.10    |            |            |    0    |   10    |       |
------------------------------------------------------------------------------------------""" \
    == str(m.get_initial_param_states())


def test_text_minos(minuit):
    assert \
r"""-------------------------------------------------
|        x        |            Valid            |
-------------------------------------------------
|      Error      |     -1.0     |      1.0     |
|      Valid      |     True     |     True     |
|    At Limit     |    False     |    False     |
|     Max FCN     |    False     |    False     |
|     New Min     |    False     |    False     |
-------------------------------------------------
-------------------------------------------------
|        y        |            Valid            |
-------------------------------------------------
|      Error      |     -0.5     |      0.5     |
|      Valid      |     True     |     True     |
|    At Limit     |    False     |    False     |
|     Max FCN     |    False     |    False     |
|     New Min     |    False     |    False     |
-------------------------------------------------""" \
    == str(minuit.minos())


def test_text_matrix(minuit):
    matrix = minuit.matrix()
    assert \
r"""-----------------
|   |    x    y |
-----------------
| x | 1.00 0.00 |
| y | 0.00 0.25 |
-----------------""" \
    == str(matrix)


def test_text_with_long_names():

    matrix = Matrix(["super-long-name", "x"], ((1.0, 0.1), (0.1, 1.0)))
    assert \
r"""-----------------------------------------------------
|                 | super-long-name               x |
-----------------------------------------------------
| super-long-name |            1.00            0.10 |
|               x |            0.10            1.00 |
-----------------------------------------------------""" \
    == str(matrix)

    mps = Params([Param(0, "super-long-name", 0, 0, False, False, False, False, False, None, None)], None)
    assert \
r"""-----------------------------------------------------------------------------------------------------
|   | Name            |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
-----------------------------------------------------------------------------------------------------
| 0 | super-long-name |     0     |     0     |            |            |         |         |       |
-----------------------------------------------------------------------------------------------------""" \
    == str(mps)


def test_console_frontend_with_difficult_values():
    matrix = Matrix(("x", "y"), ((-1.23456, 0), (0, 0)))
    assert \
r"""-----------------
|   |    x    y |
-----------------
| x | -1.2  0.0 |
| y |  0.0  0.0 |
-----------------""" \
    == str(matrix)

    mps = Params([Param(0, "x",  -1.234567e-22, 1.234567e-11, True, False, False, False, False, None, None)], None)

    assert \
r"""------------------------------------------------------------------------------------------
|   | Name |   Value   | Hesse Err | Minos Err- | Minos Err+ | Limit-  | Limit+  | Fixed |
------------------------------------------------------------------------------------------
| 0 | x    | -0.0E-11  |  1.2E-11  |            |            |         |         | CONST |
------------------------------------------------------------------------------------------""" \
    == str(mps)



from __future__ import (absolute_import, division, print_function)
import warnings
import pytest
from iminuit.tests.utils import assert_allclose
from iminuit import Minuit
from iminuit.util import Param
from iminuit.iminuit_warnings import InitialParamWarning
import numpy as np
import platform
is_pypy = platform.python_implementation()

parametrize = pytest.mark.parametrize


def test_pedantic_warning_message():
    with warnings.catch_warnings(record=True) as w:
        # use lineno of the next line for the test
        m = Minuit(lambda x: 0) # beware: this line is referred to in the test below!

        assert len(w) == 3
        for i, msg in enumerate((
            "Parameter x does not have initial value. Assume 0.",
            "Parameter x is floating but does not have initial step size. Assume 1.",
            "errordef is not given. Default to 1.")):
            assert str(w[i].message) == msg
            assert w[i].filename == __file__
            assert w[i].lineno == 18 # the lineno of "m = Minuit(lambda x: 0)"


class Func_Code:
    def __init__(self, varname):
        self.co_varnames = varname
        self.co_argcount = len(varname)


class Func1:
    def __init__(self):
        pass

    def __call__(self, x, y):
        return (x - 2.) ** 2 / 4.0 + (y - 5.) ** 2 + 10


class Func2:
    def __init__(self):
        self.func_code = Func_Code(['x', 'y'])

    def __call__(self, *arg):
        return (arg[0] - 2.) ** 2 / 4.0 + (arg[1] - 5.) ** 2 + 10


def func3(x, y):
    return (x - 2.) ** 2 / 4.0 + (y - 5.) ** 2 + 10


def func3_grad(x, y):
    dfdx = (x - 2.) / 2.0
    dfdy = 2.0 * (y - 5.)
    return [dfdx, dfdy]


def func4(x, y, z):
    return 0.2 * (x - 2.) ** 2 + 0.1 * (y - 5.) ** 2 + 0.25 * (z - 7.) ** 2 + 10


def func4_grad(x, y, z):
    dfdx = 0.4 * (x - 2.)
    dfdy = 0.2 * (y - 5.)
    dfdz = 0.5 * (z - 7.)
    return dfdx, dfdy, dfdz


def func5(x, long_variable_name_really_long_why_does_it_has_to_be_this_long, z):
    return (x ** 2) + (z ** 2) + long_variable_name_really_long_why_does_it_has_to_be_this_long ** 2


def func5_grad(x, long_variable_name_really_long_why_does_it_has_to_be_this_long, z):
    dfdx = 2 * x
    dfdy = 2 * long_variable_name_really_long_why_does_it_has_to_be_this_long
    dfdz = 2 * z
    return dfdx, dfdy, dfdz


def func6(x, m, s, a):
    return a / ((x - m) ** 2 + s ** 2)


def func7(*args):  # no signature
    return (args[0] - 1) ** 2 + (args[1] - 2) ** 2


def func8(x):  # test numpy support
    return np.sum((x - 1) ** 2)


def func8_grad(x):  # test numpy support
    return 2 * (x - 1)


class Func9:
    def __init__(self):
        sx = 2
        sy = 1
        corr = 0.5
        cov = (sx ** 2, corr * sx * sy), (corr * sx * sy, sy ** 2)
        self.cinv = np.linalg.inv(cov)

    def __call__(self, x):
        return np.dot(x.T, np.dot(self.cinv, x))


data_y = [0.552, 0.735, 0.846, 0.875, 1.059, 1.675, 1.622, 2.928,
          3.372, 2.377, 4.307, 2.784, 3.328, 2.143, 1.402, 1.44,
          1.313, 1.682, 0.886, 0.0, 0.266, 0.3]
data_x = list(range(len(data_y)))


def chi2(m, s, a):
    """Chi2 fitting routine"""
    return sum(((func6(x, m, s, a) - y) ** 2 for x, y in zip(data_x, data_y)))


def functesthelper(f, **kwds):
    m = Minuit(f, print_level=0, pedantic=False, **kwds)
    m.migrad()
    val = m.values
    assert_allclose(val['x'], 2., rtol=1e-3)
    assert_allclose(val['y'], 5., rtol=1e-3)
    assert_allclose(m.fval, 10., rtol=1e-3)
    assert m.matrix_accurate()
    assert m.migrad_ok()
    m.hesse()
    err = m.errors
    assert_allclose(err['x'], 2., rtol=1e-3)
    assert_allclose(err['y'], 1., rtol=1e-3)
    return m


def test_f1():
    functesthelper(Func1())


def test_f2():
    functesthelper(Func2())


def test_f3(): # check that providing gradient improves convergence
    m1 = functesthelper(func3)
    m2 = functesthelper(func3, grad=func3_grad)
    assert m1.get_num_call_grad() == 0
    assert m2.get_num_call_grad() > 0
    assert m1.get_num_call_fcn() > m2.get_num_call_fcn()


def test_lambda():
    functesthelper(lambda x, y: (x - 2.) ** 2 / 4.0 + (y - 5.) ** 2 + 10)


def test_nosignature():
    with pytest.raises(TypeError):
        Minuit(func7)
    m = Minuit(func7, forced_parameters=('x', 'y'),
               pedantic=False, print_level=0)
    m.migrad()
    val = m.values
    assert_allclose((val['x'], val['y'], m.fval), (1, 2, 0), atol=1e-8)
    assert m.migrad_ok()


def test_array_call():
    inf = float("infinity")
    m = Minuit(func8, a=1, b=1,
               error_a=1, error_b=1,
               limit_a=(0, inf),
               limit_b=(0, inf),
               fix_a=False,
               fix_b=False,
               print_level=0,
               errordef=1,
               use_array_call=True,
               forced_parameters=("a", "b"))
    m.migrad()
    v = m.values
    assert_allclose((v["a"], v["b"]),
                    (1, 1))
    m.hesse()
    c = m.covariance
    assert_allclose((c[("a", "a")],
                     c[("b", "b")]),
                    (1, 1))


def test_from_array_func_1():
    m = Minuit.from_array_func(func8, (2, 1),
                               error=(1, 1),
                               errordef=1,
                               print_level=0)
    assert m.fitarg == {"x0": 2,
                        "x1": 1,
                        "error_x0": 1.0,
                        "error_x1": 1.0,
                        "fix_x0": False,
                        "fix_x1": False,
                        "limit_x0": None,
                        "limit_x1": None}
    m.migrad()
    v = m.np_values()
    assert_allclose(v, (1, 1), rtol=1e-2)
    c = m.np_covariance()
    assert_allclose(np.diag(c), (1, 1), rtol=1e-2)


def test_from_array_func_2():
    m = Minuit.from_array_func(func8, (2, 1),
                               grad=func8_grad,
                               error=(0.5, 0.5),
                               limit=((0, 2), (0, 2)),
                               fix=(False, True),
                               name=("a", "b"),
                               errordef=1,
                               print_level=0)
    assert m.fitarg == {"a": 2,
                        "b": 1,
                        "error_a": 0.5,
                        "error_b": 0.5,
                        "fix_a": False,
                        "fix_b": True,
                        "limit_a": (0, 2),
                        "limit_b": (0, 2)}
    m.migrad()
    v = m.np_values()
    assert_allclose(v, (1, 1), rtol=1e-2)
    c = m.np_covariance()
    assert_allclose(c, ((1, 0), (0, 0)), rtol=1e-2)


def test_from_array_func_with_broadcasting():
    m = Minuit.from_array_func(func8, (1, 1),
                               error=0.5,
                               limit=(0, 2),
                               errordef=1,
                               print_level=0)
    assert m.fitarg == {"x0": 1,
                        "x1": 1,
                        "error_x0": 0.5,
                        "error_x1": 0.5,
                        "fix_x0": False,
                        "fix_x1": False,
                        "limit_x0": (0, 2),
                        "limit_x1": (0, 2)}
    m.migrad()
    v = m.np_values()
    assert_allclose(v, (1, 1))
    c = m.np_covariance()
    assert_allclose(np.diag(c), (1, 1))


def test_view_repr():
    m = Minuit(func3, print_level=0, errordef=1,
               x=1, y=2, error_x=3, error_y=4)
    mid = id(m)
    assert repr(m.values) == ("""
<ValueView of Minuit at %x>
  x: 1.0
  y: 2.0
""" % mid).strip()
    assert repr(m.args) == ("""
<ArgsView of Minuit at %x>
  1.0
  2.0
""" % mid).strip()


def test_no_resume():
    m = Minuit(func3, print_level=0, pedantic=False)
    m.migrad()
    n = m.get_num_call_fcn()
    m.migrad()
    assert m.get_num_call_fcn() > n
    m.migrad(resume=False)
    assert m.get_num_call_fcn() == n

    m = Minuit(func3, grad=func3_grad, print_level=0, pedantic=False)
    m.migrad()
    n = m.get_num_call_fcn()
    k = m.get_num_call_grad()
    m.migrad()
    assert m.get_num_call_fcn() > n
    assert m.get_num_call_grad() > k
    m.migrad(resume=False)
    assert m.get_num_call_fcn() == n
    assert m.get_num_call_grad() == k


def test_typo():
    with pytest.raises(RuntimeError):
        Minuit(func4, printlevel=0)


def test_non_invertible():
    # making sure it doesn't crash
    def f(x, y):
        return (x * y) ** 2

    m = Minuit(f, pedantic=False, print_level=0)
    m.migrad()
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        try:
            m.hesse()
            raise RuntimeError('Hesse did not raise a warning')
        except Warning:
            pass
    try:
        m.matrix()
        raise RuntimeError('Matrix did not raise an error')  # shouldn't reach here
    except RuntimeError:
        pass


@parametrize('grad', (None, func3_grad))
def test_fix_param(grad):
    kwds = {'print_level': 0, 'pedantic': False, 'grad': grad}
    m = Minuit(func3, **kwds)
    m.migrad()
    m.minos()
    assert_allclose(m.np_values(), (2, 5), rtol=1e-2)
    assert_allclose(m.np_errors(), (2, 1))
    assert_allclose(m.matrix(), ((4, 0), (0, 1)), atol=1e-4)
    for b in (True, False):
        assert_allclose(m.matrix(skip_fixed=b), [[4, 0], [0, 1]], atol=1e-4)
    # now fix y = 10
    m = Minuit(func3, y=10., fix_y=True, **kwds)
    m.migrad()
    assert_allclose(m.np_values(), (2, 10), rtol=1e-2)
    assert_allclose(m.fval, 35)
    assert m.list_of_vary_param() == ['x']
    assert m.list_of_fixed_param() == ['y']
    assert_allclose(m.matrix(skip_fixed=True), [[4]], atol=1e-4)
    assert_allclose(m.matrix(skip_fixed=False), [[4, 0], [0, 0]], atol=1e-4)

    assert m.is_fixed('x') is False
    assert m.is_fixed('y') is True

    m.fixed['x'] = True
    m.fixed['y'] = False
    m.migrad()
    m.hesse()
    assert_allclose(m.np_values(), (2, 5), rtol=1e-2)
    assert_allclose(m.matrix(skip_fixed=True), [[1]], atol=1e-4)
    assert_allclose(m.matrix(skip_fixed=False), [[0, 0], [0, 1]], atol=1e-4)

    with pytest.raises(KeyError):
        m.is_fixed('a')

    # fix by setting limits
    m = Minuit(func3, y=10., limit_y=(10, 10), pedantic=False, print_level=0)
    assert m.fixed['y']

    # initial value out of range is forced in range
    m = Minuit(func3, y=20., limit_y=(10, 10), pedantic=False, print_level=0)
    assert m.fixed['y']
    assert m.values['y'] == 10


def test_fitarg_oneside():
    m = Minuit(func4, print_level=0, y=10., fix_y=True, limit_x=(None, 20.),
               pedantic=False)
    fitarg = m.fitarg
    assert_allclose(fitarg['y'], 10.)
    assert fitarg['fix_y']
    assert fitarg['limit_x'] == (None, 20)
    m.migrad()

    fitarg = m.fitarg

    assert_allclose(fitarg['x'], 2., atol=1e-2)
    assert_allclose(fitarg['y'], 10., atol=1e-2)
    assert_allclose(fitarg['z'], 7., atol=1e-2)

    assert 'error_y' in fitarg
    assert 'error_x' in fitarg
    assert 'error_z' in fitarg

    assert fitarg['fix_y']
    assert fitarg['limit_x'] == (None, 20)


def test_fitarg():
    m = Minuit(func4, print_level=-1, y=10., fix_y=True, limit_x=(0, 20.),
               pedantic=False)
    fitarg = m.fitarg
    assert_allclose(fitarg['y'], 10.)
    assert fitarg['fix_y'] is True
    assert fitarg['limit_x'] == (0, 20)
    m.migrad()

    fitarg = m.fitarg

    assert_allclose(fitarg['y'], 10.)
    assert_allclose(fitarg['x'], 2., atol=1e-2)
    assert_allclose(fitarg['z'], 7., atol=1e-2)

    assert 'error_y' in fitarg
    assert 'error_x' in fitarg
    assert 'error_z' in fitarg

    assert fitarg['fix_y'] is True
    assert fitarg['limit_x'] == (0, 20)


@parametrize('grad', (None, func3_grad))
@parametrize('sigma', (1, 4))
def test_minos_all(grad, sigma):
    m = Minuit(func3, grad=func3_grad, pedantic=False, print_level=0)
    m.migrad()
    m.minos(sigma=sigma)
    assert_allclose(m.merrors[('x', -1.0)], -sigma * 2, rtol=1e-2)
    assert_allclose(m.merrors[('x', 1.0)], sigma * 2, rtol=1e-2)
    assert_allclose(m.merrors[('y', 1.0)], sigma * 1, rtol=1e-2)


@parametrize('grad', (None, func3_grad))
def test_minos_single(grad):
    m = Minuit(func3, grad=func3_grad, pedantic=False, print_level=0)
    m.set_strategy(2)
    m.migrad()
    m.minos('x')
    assert_allclose(m.merrors[('x', -1.0)], -2)
    assert_allclose(m.merrors[('x', 1.0)], 2)


def test_minos_single_fixed_raising():
    m = Minuit(func3, pedantic=False, print_level=0, fix_x=True)
    m.migrad()
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        with pytest.raises(RuntimeWarning):
            m.minos('x')

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        ret = m.minos('x')
        assert ret is None


def test_minos_single_no_migrad():
    m = Minuit(func3, pedantic=False, print_level=0)
    with pytest.raises(RuntimeError):
        m.minos('x')


def test_minos_single_nonsense_variable():
    m = Minuit(func3, pedantic=False, print_level=0)
    m.migrad()
    with pytest.raises(RuntimeError):
        m.minos('nonsense')


@parametrize('grad', (None, func5_grad))
def test_fixing_long_variable_name(grad):
    m = Minuit(func5, grad=grad, pedantic=False, print_level=0,
               fix_long_variable_name_really_long_why_does_it_has_to_be_this_long=True,
               long_variable_name_really_long_why_does_it_has_to_be_this_long=0)
    m.migrad()


def test_initial_value():
    m = Minuit(func3, pedantic=False, x=1., y=2., error_x=3., print_level=0)
    assert_allclose(m.args[0], 1.)
    assert_allclose(m.args[1], 2.)
    assert_allclose(m.values['x'], 1.)
    assert_allclose(m.values['y'], 2.)
    assert_allclose(m.errors['x'], 3.)


@parametrize('grad', (None, func3_grad))
@parametrize('sigma', (1, 2))
def test_mncontour(grad, sigma):
    m = Minuit(func3, grad=grad, pedantic=False,
               x=1., y=2., error_x=3., print_level=0)
    m.migrad()
    xminos, yminos, ctr = m.mncontour('x', 'y', numpoints=30, sigma=sigma)
    xminos_t = m.minos('x', sigma=sigma)['x']
    yminos_t = m.minos('y', sigma=sigma)['y']
    assert_allclose(xminos.upper, xminos_t.upper)
    assert_allclose(xminos.lower, xminos_t.lower)
    assert_allclose(yminos.upper, yminos_t.upper)
    assert_allclose(yminos.lower, yminos_t.lower)
    assert len(ctr) == 30
    assert len(ctr[0]) == 2


@parametrize('grad', (None, func3_grad))
def test_contour(grad):
    # FIXME: check the result
    m = Minuit(func3, grad=grad, pedantic=False,
               x=1., y=2., error_x=3., print_level=0)
    m.migrad()
    m.contour('x', 'y')


@parametrize('grad', (None, func3_grad))
def test_profile(grad):
    # FIXME: check the result
    m = Minuit(func3, grad=grad, pedantic=False,
               x=1., y=2., error_x=3., print_level=0)
    m.migrad()
    m.profile('y')


@parametrize('grad', (None, func3_grad))
def test_mnprofile(grad):
    # FIXME: check the result
    m = Minuit(func3, grad=grad, pedantic=False, x=1., y=2., error_x=3., print_level=0)
    m.migrad()
    m.mnprofile('y')


def test_printfmin_uninitialized():
    # issue 85
    def f(x):
        return 2 + 3 * x

    fitter = Minuit(f, pedantic=False)
    with pytest.raises(RuntimeError):
        fitter.print_fmin()


def test_reverse_limit():
    # issue 94
    def f(x, y, z):
        return (x - 2) ** 2 + (y - 3) ** 2 + (z - 4) ** 2

    with pytest.raises(ValueError):
        Minuit(f, limit_x=(3., 2.), pedantic=False, print_level=0)


class TestOutputInterface:
    def setup(self):
        self.m = Minuit(func3, print_level=0, pedantic=False)
        self.m.migrad()
        self.m.hesse()
        self.m.minos()

    def test_args(self):
        actual = self.m.args
        expected = [2., 5.]
        assert_allclose(actual, expected, atol=1e-8)

    def test_matrix(self):
        actual = self.m.matrix()
        expected = [[4., 0.], [0., 1.]]
        assert_allclose(actual, expected, atol=1e-8)

    def test_matrix_correlation(self):
        actual = self.m.matrix(correlation=True)
        expected = [[1., 0.], [0., 1.]]
        assert_allclose(actual, expected, atol=1e-8)

    def test_np_matrix(self):
        actual = self.m.np_matrix()
        expected = [[4., 0.], [0., 1.]]
        assert_allclose(actual, expected, atol=1e-8)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2, 2)

    def test_np_matrix_correlation(self):
        actual = self.m.np_matrix(correlation=True)
        expected = [[1., 0.], [0., 1.]]
        assert_allclose(actual, expected, atol=1e-8)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2, 2)

    def test_np_values(self):
        actual = self.m.np_values()
        expected = [2., 5.]
        assert_allclose(actual, expected, atol=1e-8)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2,)

    def test_np_errors(self):
        actual = self.m.np_errors()
        expected = [2., 1.]
        assert_allclose(actual, expected, atol=1e-8)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2,)

    def test_np_merrors(self):
        actual = self.m.np_merrors()
        # output format is [abs(down_delta), up_delta] following
        # the matplotlib convention in matplotlib.pyplot.errorbar
        down_delta = (-2, -1)
        up_delta = (2, 1)
        assert_allclose(actual, (np.abs(down_delta), up_delta), atol=1e-8)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2, 2)

    def test_np_covariance(self):
        actual = self.m.np_covariance()
        expected = [[4., 0.], [0., 1.]]
        assert_allclose(actual, expected, atol=1e-8)
        assert isinstance(actual, np.ndarray)
        assert actual.shape == (2, 2)


def test_chi2_fit():
    """Fit a curve to data."""
    m = Minuit(chi2, s=2., error_a=0.1, errordef=0.01,
               print_level=0, pedantic=False)
    m.migrad()
    output = [round(10 * m.values['a']), round(100 * m.values['s']),
              round(100 * m.values['m'])]
    expected = [round(10 * 64.375993), round(100 * 4.267970),
                round(100 * 9.839172)]
    assert_allclose(output, expected)


def test_oneside():
    m_limit = Minuit(func3, limit_x=(None, 9), pedantic=False, print_level=0)
    m_nolimit = Minuit(func3, pedantic=False, print_level=0)
    # Solution: x=2., y=5.
    m_limit.tol = 1e-4
    m_nolimit.tol = 1e-4
    m_limit.migrad()
    m_nolimit.migrad()
    assert_allclose(list(m_limit.values.values()),
                    list(m_nolimit.values.values()), atol=1e-4)


def test_oneside_outside():
    m = Minuit(func3, limit_x=(None, 1), pedantic=False, print_level=0)
    m.migrad()
    assert_allclose(m.values['x'], 1)


def test_num_call():
    class Func:
        ncall = 0

        def __call__(self, x):
            self.ncall += 1
            return x ** 2

    # check that counting is accurate
    func = Func()
    m = Minuit(func, pedantic=False, print_level=0)
    m.migrad()
    assert m.get_num_call_fcn() == func.ncall
    m.migrad()
    assert m.get_num_call_fcn() == func.ncall
    func.ncall = 0
    m.migrad(resume=False)
    assert func.ncall == m.get_num_call_fcn()

    ncall_without_limit = m.get_num_call_fcn()
    # check that ncall argument limits function calls in migrad
    # note1: Minuit only checks the ncall counter in units of one iteration
    # step, therefore the call counter is in general not equal to ncall.
    # note2: If you pass ncall=0, Minuit uses a heuristic value that depends
    # on the number of parameters.
    m.migrad(ncall=1, resume=False)
    assert m.get_num_call_fcn() < ncall_without_limit


def test_set_error_def():
    m = Minuit(lambda x: x ** 2, pedantic=False, print_level=0, errordef=1)
    m.migrad()
    m.hesse()
    assert_allclose(m.errors["x"], 1)
    m.set_errordef(4)
    m.hesse()
    assert_allclose(m.errors["x"], 2)


def test_get_param_states():
    m = Minuit(func3, x=1, y=2, error_x=3, error_y=4,
               fix_x=True, limit_y=(None, 10),
               pedantic=False, errordef=1, print_level=0)
    # these are the initial param states
    expected = [Param(0, 'x', 1.0, 3.0, False, True, False, False, False, None, None),
                Param(1, 'y', 2.0, 4.0, False, False, True, False, True, None, 10)]
    assert m.get_param_states() == expected

    m.migrad()
    assert m.get_initial_param_states() == expected

    expected = [Param(0, 'x', 1.0, 3.0, False, True, False, False, False, None, None),
                Param(1, 'y', 5.0, 1.0, False, False, True, False, True, None, 10)]

    params = m.get_param_states()
    for i, exp in enumerate(expected):
        p = params[i]
        assert set(p._fields) == set(exp._fields)
        for key in exp._fields:
            if key in ('value', 'error'):
                assert_allclose(getattr(p, key), getattr(exp, key), rtol=1e-2)
            else:
                assert getattr(p, key) == getattr(exp, key)


def test_latex_matrix():
    m = Minuit.from_array_func(Func9(), (0, 0), name=('x', 'y'),
                               pedantic=False, print_level=0)
    m.migrad()
    assert r"""%\usepackage[table]{xcolor} % include this for color
%\usepackage{rotating} % include this for rotate header
%\documentclass[xcolor=table]{beamer} % for beamer
\begin{tabular}{|c|c|c|}
\hline
\rotatebox{90}{} & \rotatebox{90}{x} & \rotatebox{90}{y}\\
\hline
x & \cellcolor[RGB]{255,118,118} 1 & \cellcolor[RGB]{209,186,152} 0.5\\
\hline
y & \cellcolor[RGB]{209,186,152} 0.5 & \cellcolor[RGB]{255,118,118} 1\\
\hline
\end{tabular}""" == str(m.latex_matrix())


def test_non_analytical_function():
    class Func:
        i = 0

        def __call__(self, a):
            self.i += 1
            return self.i % 3

    m = Minuit(Func(), pedantic=False, print_level=0)
    fmin, param = m.migrad()
    assert fmin.is_valid is False
    assert fmin.is_above_max_edm is True


def test_function_without_local_minimum():
    m = Minuit(lambda a: -a, pedantic=False, print_level=0)
    fmin, param = m.migrad()
    assert fmin.is_valid is False
    assert fmin.is_above_max_edm is True


# Bug in Minuit2, this needs to be fixed in upstream ROOT
@pytest.mark.xfail(strict=True)
def test_function_with_maximum():
    def func(a):
        return -a ** 2

    m = Minuit(func, pedantic=False, print_level=0)
    fmin, param = m.migrad()
    assert fmin.is_valid is False


def test_perfect_correlation():
    def func(a, b):
        return (a - b) ** 2

    m = Minuit(func, pedantic=False, print_level=0)
    fmin, param = m.migrad()
    assert fmin.is_valid is True
    assert fmin.has_accurate_covar is False
    assert fmin.has_posdef_covar is False
    assert fmin.has_made_posdef_covar is True


def test_modify_param_state():
    m = Minuit(func3, x=1, y=2, fix_y=True, pedantic=False,
               print_level=0)
    m.migrad()
    assert_allclose(m.np_values(), [2, 2], atol=1e-2)
    assert_allclose(m.np_errors(), [2, 1], atol=1e-2)
    m.fixed['y'] = False
    m.values['x'] = 1
    m.errors['x'] = 1
    assert_allclose(m.np_values(), [1, 2], atol=1e-2)
    assert_allclose(m.np_errors(), [1, 1], atol=1e-2)
    m.migrad()
    assert_allclose(m.np_values(), [2, 5], atol=1e-2)
    assert_allclose(m.np_errors(), [2, 1], atol=1e-2)
    m.values['y'] = 6
    m.hesse()  # does not change minimum
    assert_allclose(m.np_values(), [2, 6], atol=1e-2)
    assert_allclose(m.np_errors(), [2, 1], atol=1e-2)


def test_view_lifetime():
    m = Minuit(func3, x=1, y=2, pedantic=False, print_level=0)
    val = m.values
    arg = m.args
    del m
    val['x'] = 3  # should not segfault
    assert val['x'] == 3
    arg[0] = 5  # should not segfault
    assert arg[0] == 5


def test_bad_functions():
    def throwing(x):
        raise RuntimeError("user message")
    divide_by_zero = lambda x: 1/0
    returning_nan = lambda x: np.nan
    returning_garbage = lambda x: "foo"

    for func, expected in (
            (throwing, 'RuntimeError("user message")'),
            (divide_by_zero, "ZeroDivisionError"),
            (returning_nan, "result is NaN"),
            (returning_garbage, "TypeError")):
        m = Minuit(func, x=1, pedantic=False, throw_nan=True,
                   print_level=0)
        with pytest.raises(RuntimeError) as excinfo:
            m.migrad()
        assert expected in excinfo.value.args[0]

    returning_nan = lambda x: np.array([1, np.nan])
    returning_noniterable = lambda x: 0
    returning_garbage = lambda x: np.array([1, "foo"])

    for func, expected in (
            (throwing, 'RuntimeError("user message")'),
            (divide_by_zero, "ZeroDivisionError"),
            (returning_nan, "result is NaN"),
            (returning_garbage, "TypeError"),
            (returning_noniterable, "TypeError")):
        m = Minuit.from_array_func(lambda x: 0, (1, 1), grad=func,
                                   pedantic=False, throw_nan=True,
                                   print_level=0)
        with pytest.raises(RuntimeError) as excinfo:
            m.migrad()
        if is_pypy and func is returning_garbage:
            pass
        else:
            assert expected in excinfo.value.args[0]



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__all__ = ['Gradient']


class Gradient:
    """Color gradient.
    """
    # from http://code.activestate.com/recipes/266466-html-colors-tofrom-rgb-tuples/
    @classmethod
    def color_for(cls, v, min=0., max=1., startcolor=(163, 254, 186),
                  stopcolor=(255, 118, 118)):
        c = [0] * 3
        for i, sc in enumerate(startcolor):
            offset = 1.0 * (v - min) * (stopcolor[i] - startcolor[i]) / (max - min)
            c[i] = round(startcolor[i] + offset)
        return tuple(c)

    @classmethod
    def xcolor_for(cls, v):  # pragma: no cover
        c = cls.color_for(abs(v))
        return '[rgb]{%3.2f,%3.2f,%3.2f}' % (c[0] / 255., c[1] / 255., c[2] / 255.)

    @classmethod
    def rgb_color_for(cls, v):
        c = cls.color_for(abs(v))
        return 'rgb(%d,%d,%d)' % c



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from iminuit.color import Gradient

__all__ = ['LatexTable']


class LatexTable:
    """Latex table output.
    """
    float_format = '%10.5g'
    int_format = '%d'
    latex_kwd = [
        'alpha', 'beta', 'gamma',
        'delta', 'epsilon', 'zeta',
        'eta', 'theta', 'iota',
        'kappa', 'lambda', 'mu',
        'nu', 'xi', 'omicron',
        'pi', 'rho', 'sigma',
        'tau', 'upsilon', 'phi',
        'chi', 'psi', 'omega',
        'Alpha', 'Beta', 'Gamma',
        'Delta', 'Epsilon', 'Zeta',
        'Eta', 'Theta', 'Iota',
        'Kappa', 'Lambda', 'Mu',
        'Nu', 'Xi', 'Omicron',
        'Pi', 'Rho', 'Sigma',
        'Tau', 'Upsilon', 'Phi',
        'Chi', 'Psi', 'Omega',
    ]

    def __init__(self, data, headers=None, smart_latex=True,
                 escape_under_score=True, alignment=None, rotate_header=False,
                 latex_map=None):
        # Make sure #data columns matches #header columns, using any non-zero
        # length if data or headers are missing
        if len(data) > 0:
            num_col = len(data[0])
            if headers:
                assert num_col == len(headers)
        else:
            if headers is not None:
                num_col = len(headers)
            else:
                # LaTeX requires at least one column
                num_col = 1

        self.headers = headers
        self.data = data
        self.num_col = num_col
        self.smart_latex = smart_latex
        self.escape_under_score = escape_under_score
        self.alignment = self._auto_align() if alignment is None else alignment
        self.rotate_header = rotate_header
        self.latex_map = {} if latex_map is None else latex_map
        self.cell_color = {}  # map of tuple (i,j)=>(r, g, b) #i,j include header

    def _auto_align(self):
        return '|' + 'c|' * self.num_col

    def _format(self, s):
        if s in self.latex_map:
            return self.latex_map[s]
        elif isinstance(s, float):
            return self.float_format % s
        elif isinstance(s, int):
            return self.int_format % s
        elif self.smart_latex:
            return self._convert_smart_latex(s)
        elif self.escape_under_score:
            return s.replace('_', r'\_')
        else:
            return s

    def _convert_smart_latex(self, s):
        """convert greek symbol to latex one
        transform
        a to $a$ if a is greek letter else just a
        a_xxx to $a_{xxx}$ and
        a_xxx_yyy_zzz to a xxx $yyy_{zzz}$
        """
        # FIXME: implement this

        parts = s.split('_')
        if len(parts) == 1:  # a to $a$ if a is greek letter else just a
            if parts[0] in self.latex_kwd:
                return r'$\%s$' % str(parts[0])
            else:
                return str(parts[0])
        elif len(parts) == 2:  # a_xxx to $a_{xxx}$ and
            first = '\\%s' % parts[0] if parts[0] in self.latex_kwd else parts[0]
            second = '\\%s' % parts[1] if parts[1] in self.latex_kwd else parts[1]
            return r'$%s_{%s}$' % (first, second)
        else:  # a_xxx_yyy_zzz to a xxx $yyy_{zzz}$
            textpart = map(self._convert_smart_latex, parts[:-2])
            textpart = ' '.join(textpart)
            latexpart = self._convert_smart_latex('_'.join(parts[-2:]))
            return textpart + ' ' + latexpart

    def set_cell_color(self, i, j, c):
        """colorize i,j cell with rgb color tuple c

        Note that i,j index includes header.
        i=0 is header if header is present. If header is not present then
        i=0 refer to first data row.
        """
        self.cell_color[(i, j)] = c

    def _prepare(self):  # return list of list
        ret = []
        if self.headers:
            tmp = list(map(self._format, self.headers))
            if self.rotate_header:
                tmp = list(map(lambda x: '\\rotatebox{90}{%s}' % x, tmp))

            ret.append(tmp)
        for x in self.data:
            ret.append(list(map(self._format, x)))
        return ret

    def __str__(self):
        hline = '\\hline\n'
        ret = ''
        if len(self.cell_color) != 0:
            ret += '%\\usepackage[table]{xcolor} % include this for color\n'
            ret += '%\\usepackage{rotating} % include this for rotate header\n'
            ret += '%\\documentclass[xcolor=table]{beamer} % for beamer\n'
        ret += '\\begin{tabular}{%s}\n' % self.alignment
        ret += hline
        tdata = self._prepare()
        # decorate it

        for (i, j), c in self.cell_color.items():
            xcolor = '[RGB]{%d,%d,%d}' % (c[0], c[1], c[2])
            tdata[i][j] = '\\cellcolor' + xcolor + ' ' + tdata[i][j]

        for line in tdata:
            ret += ' & '.join(line) + '\\\\\n'
            ret += hline
        ret += '\\end{tabular}\n'

        return ret.strip()


class LatexFactory:
    @classmethod
    def build_matrix(cls, vnames, matrix, latex_map=None):
        """build latex correlation matrix"""
        # ret_link  = '<a onclick="$(\'#%s\').toggle()" href="#">Show Latex</a>'%uid
        headers = [''] + list(vnames)
        data = []
        color = {}
        for i, v1 in enumerate(vnames):
            tmp = [v1]
            for j, v2 in enumerate(vnames):
                m = matrix[i][j]
                tmp.append(m)
                color[(i + 1, j + 1)] = Gradient.color_for(abs(m))
                # +1 for header on the side and top
            data.append(tmp)

        table = LatexTable(headers=headers, data=data, rotate_header=True,
                           latex_map=latex_map)
        table.float_format = '%.2g'
        for (i, j), c in color.items():
            table.set_cell_color(i, j, c)
        return table

    @classmethod
    def build_param_table(cls, mps, merr=None, float_format='%5.3e',
                          smart_latex=True, latex_map=None):
        """build latex parameter table"""
        headers = ['', 'Name', 'Value', 'Hesse Error', 'Minos Error-',
                   'Minos Error+', 'Limit-', 'Limit+', 'Fixed?', ]

        data = []
        for i, mp in enumerate(mps):
            minos_p, minos_m = ('', '') if merr is None or mp.name not in merr else \
                ('%g' % merr[mp.name].upper, '%g' % merr[mp.name].lower)
            limit_p = '' if mp.upper_limit is None else '%g' % mp.upper_limit
            limit_m = '' if mp.lower_limit is None else '%s' % mp.lower_limit
            fixed = 'Yes' if mp.is_fixed else 'No'
            tmp = [
                i,
                mp.name,
                '%g' % mp.value,
                '%g' % mp.error,
                minos_m,
                minos_p,
                limit_m,
                limit_p,
                fixed,
            ]
            data.append(tmp)
        alignment = '|c|r|r|r|r|r|r|r|c|'
        ret = LatexTable(data, headers=headers, alignment=alignment,
                         smart_latex=smart_latex, latex_map=latex_map)
        ret.float_format = float_format
        return ret



from __future__ import (absolute_import, division, print_function)
from iminuit.util import (fitarg_rename,
                          true_param,
                          param_name,
                          extract_iv,
                          extract_limit,
                          extract_error,
                          extract_fix,
                          remove_var,
                          arguments_from_docstring,
                          Matrix,
                          FMin,
                          Param,
                          MError,
                          Params,
                          MigradResult)
import pytest


def test_fitarg_rename():
    fitarg = {'x': 1, 'limit_x': (2, 3), 'fix_x': True, 'error_x': 10}

    def ren(x):
        return 'z_' + x

    newfa = fitarg_rename(fitarg, ren)
    assert 'z_x' in newfa
    assert 'limit_z_x' in newfa
    assert 'error_z_x' in newfa
    assert 'fix_z_x' in newfa
    assert len(newfa) == 4


def test_fitarg_rename_strprefix():
    fitarg = {'x': 1, 'limit_x': (2, 3), 'fix_x': True, 'error_x': 10}
    newfa = fitarg_rename(fitarg, 'z')
    assert 'z_x' in newfa
    assert 'limit_z_x' in newfa
    assert 'error_z_x' in newfa
    assert 'fix_z_x' in newfa
    assert len(newfa) == 4


def test_true_param():
    assert true_param('N') is True
    assert true_param('limit_N') is False
    assert true_param('error_N') is False
    assert true_param('fix_N') is False


def test_param_name():
    assert param_name('N') == 'N'
    assert param_name('limit_N') == 'N'
    assert param_name('error_N') == 'N'
    assert param_name('fix_N') == 'N'


def test_extract_iv():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_iv(d)
    assert 'k' in ret
    assert 'limit_k' not in ret
    assert 'error_k' not in ret
    assert 'fix_k' not in ret


def test_extract_limit():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_limit(d)
    assert 'k' not in ret
    assert 'limit_k' in ret
    assert 'error_k' not in ret
    assert 'fix_k' not in ret


def test_extract_error():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_error(d)
    assert 'k' not in ret
    assert 'limit_k' not in ret
    assert 'error_k' in ret
    assert 'fix_k' not in ret


def test_extract_fix():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_fix(d)
    assert 'k' not in ret
    assert 'limit_k' not in ret
    assert 'error_k' not in ret
    assert 'fix_k' in ret


def test_remove_var():
    dk = dict(k=1, limit_k=1, error_k=1, fix_k=1)
    dl = dict(l=1, limit_l=1, error_l=1, fix_l=1)
    dm = dict(m=1, limit_m=1, error_m=1, fix_m=1)
    dn = dict(n=1, limit_n=1, error_n=1, fix_n=1)
    d = {}
    d.update(dk)
    d.update(dl)
    d.update(dm)
    d.update(dn)

    ret = remove_var(d, ['k', 'm'])
    for k in dk:
        assert k not in ret
    for k in dl:
        assert k in ret
    for k in dm:
        assert k not in ret
    for k in dn:
        assert k in ret


def test_arguments_from_docstring():
    s = 'f(x, y, z)'
    a = arguments_from_docstring(s)
    assert a == ['x', 'y', 'z']
    # this is a hard one
    s = 'Minuit.migrad( int ncall_me =10000, [resume=True, int nsplit=1])'
    a = arguments_from_docstring(s)
    assert a == ['ncall_me', 'resume', 'nsplit']


def test_Matrix():
    x = Matrix(("a", "b"), [[1, 2],[3, 4]])
    assert x[0] == (1, 2)
    assert x[1] == (3, 4)
    assert x == ((1, 2), (3, 4))
    assert repr(x) == '((1, 2), (3, 4))'
    with pytest.raises(TypeError):
        x[0] = (1, 2)
    with pytest.raises(TypeError):
        x[0][0] = 1


def test_Param():
    keys = "number name value error is_const is_fixed has_limits has_lower_limit has_upper_limit lower_limit upper_limit".split()
    values = 3, "foo", 1.2, 3.4, False, False, True, True, False, 42, None
    p = Param(*values)

    assert p.has_lower_limit == True
    assert p.has_upper_limit == False
    assert p["has_lower_limit"] == True
    assert p["lower_limit"] == 42
    assert p["upper_limit"] == None
    assert "upper_limit" in p
    assert "foo" not in p

    assert [key for key in p] == keys
    assert p.keys() == tuple(keys)
    assert p.values() == values
    assert p.items() == tuple((k, v) for k, v in zip(keys, values))

    assert str(p) == "Param(number=3, name='foo', value=1.2, error=3.4, is_const=False, is_fixed=False, has_limits=True, has_lower_limit=True, has_upper_limit=False, lower_limit=42, upper_limit=None)"


def test_MError():
    keys = "name is_valid lower upper lower_valid upper_valid at_lower_limit at_upper_limit at_lower_max_fcn at_upper_max_fcn lower_new_min upper_new_min nfcn min".split()
    values = "Foo", True, 0.1, 1.2, True, True, False, False, False, False, 0.1, 1.2, 42, 0.2

    m = MError(*values)

    assert m.keys() == tuple(keys)
    assert m.values() == values
    assert m.name == "Foo"
    assert m["name"] == "Foo"


def test_FMin():
    keys = "fval edm tolerance nfcn ncalls up is_valid has_valid_parameters has_accurate_covar has_posdef_covar has_made_posdef_covar hesse_failed has_covariance is_above_max_edm has_reached_call_limit".split()
    values = 0.2, 1e-3, 0.1, 10, 10, 1.2, False, False, False, False, False, False, False, False, False

    f = FMin(*values)

    assert f.keys() == tuple(keys)
    assert f.values() == values
    assert f.fval == 0.2
    assert f["fval"] == 0.2


def test_MigradResult():
    fmin = FMin(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14)
    params = Params([], None)
    mr = MigradResult(fmin, params)
    assert mr.fmin is fmin
    assert mr[0] is fmin
    assert mr.params is params
    assert mr[1] is params
    a, b = mr
    assert a is fmin
    assert b is params
